package ntsP2P;

import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.net.Socket;
import java.net.UnknownHostException;
import java.util.Hashtable;
import java.util.Vector;

import monitor.CPUMonitor;
import monitor.LatencyMonitor;
import salsa.language.Message;
import salsa.language.ServiceFactory;
import salsa.language.UniversalActor;
import salsa.naming.UAN;
import ats.ActorProfile;
import core.agents.ProfilingAgent;


public class NTSProfilingAgent implements ProfilingAgent {

	private LatencyMonitor		latencyMonitor = null;
	private	CPUMonitor		cpuMonitor = null;

	private double			threshold = 0.3;
	private	double			maxFLOPS = 0;
	public	double			getMaxFLOPS()			{ return maxFLOPS; }
	public	boolean 		isLight()			{ return getAvailableProcessing() > (maxFLOPS*threshold); }

	private	long			lastResetTime = System.currentTimeMillis();
	private long			timeSinceLastReset()		{ return System.currentTimeMillis() - lastResetTime; }

	private	boolean			computing = true;
	public	boolean			isComputing()			{ return computing; }
	public	void			setComputing(boolean computing)	{ this.computing = computing; }

	public NTSProfilingAgent() {
		if ( System.getProperty("silent") == null) {
			System.err.println("Determining cpu performance:");
		}

                long startTime = System.currentTimeMillis();
                long l = 2, calculations = 300000;
                for (int i = 0; i < calculations; i++) l *= 2;
                maxFLOPS = calculations / (double)((double)(System.currentTimeMillis()-startTime)/1000.0);

		if ( System.getProperty("silent") == null) {
			System.err.println("CPU performance: " + maxFLOPS*2 + " flops");
		}

		maxFLOPS *= 2;
		//typically the real max is about twice the inital FLOPS reading

		/**
		 *	Create the monitoring services for the profiling agent:
		 *	the latency and the cpu monitors
		 */
		latencyMonitor = new LatencyMonitor();
		latencyMonitor.start();

		/**
		 *	Set the latency to ourself to 0
		 */
		peersToLatency.put( ServiceFactory.getTheater().getHost(), new Double(0) );
	}

	private Hashtable actorProfiles = new Hashtable();
	private long totalProcessed = 0;
	private long processedSinceStart = 0;

	public double getAvailableProcessing() {
		Vector profiles = getActorProfiles();

		double usedFLOPS = 0;
		for (int i = 0; i < profiles.size(); i++) {
			ActorProfile candidate = (ActorProfile)profiles.get(i);

			if (candidate.timeProcessing() > candidate.timeSending()) {
				usedFLOPS += maxFLOPS*((double)(candidate.timeProcessing()-candidate.timeSending())/timeSinceLastReset());
			}
		}
		return maxFLOPS-usedFLOPS;
	}

	public double getFLOPSUsedBy(ActorProfile candidate) {
		Vector profiles = getActorProfiles();

		double totalProcessingTime = 0;
		for (int i = 0; i < profiles.size(); i++) {
			ActorProfile current = (ActorProfile)profiles.get(i);

			if (current.timeProcessing() > current.timeSending()) {
				totalProcessingTime += (double)(current.timeProcessing()-current.timeSending())/(double)timeSinceLastReset();
			}
		}
		double candidateTime = 0;
		if (candidate.timeProcessing() > candidate.timeSending()) {
			candidateTime = (double)(candidate.timeProcessing()-candidate.timeSending())/(double)timeSinceLastReset();
		}
		return maxFLOPS * (candidateTime/totalProcessingTime);
	}

	Vector stealSources = new Vector();

	public int getDistinctSteals() {
		return stealSources.size();
	}

	public void gotStealFrom(String source) {
		if (!stealSources.contains(source)) stealSources.add(source);
	}


	public long processed()				{ return totalProcessed; }

	public Vector getActorProfiles() 		{ return new Vector(actorProfiles.values()); }
	public Vector getActorReferences() {
		Vector profiles = getActorProfiles();
		Vector references = new Vector();

		for (int i = 0; i < profiles.size(); i++) {
			references.add( UniversalActor.getReferenceByName( ((ActorProfile)profiles.get(i)).getUAN()) );
		}
		return references;
	}

	public void beginProcessed(UAN identifier, Message message) {
		totalProcessed++;
		processedSinceStart++;

		ActorProfile target = (ActorProfile)actorProfiles.get(identifier);
		if (target == null) return;  
    
		target.processedMessage();
		target.beginProcessing();
	}
	public void endProcessed(UAN identifier, Message message) { 
		ActorProfile target = (ActorProfile)actorProfiles.get(identifier);
		if (target == null) return;  
		target.endProcessing();
	}

	public void received(UAN identifier, Message message) {
		if (message.getSource().getUAN() != null) {
			ActorProfile target = (ActorProfile)actorProfiles.get(message.getSource().getUAN());
			if (target == null) return;
			else if (message.getTarget().getUAL() != null) {
				target.incrementSent( message.getTarget().getUAL().getHost() );    
			}
		}

		ActorProfile target = (ActorProfile)actorProfiles.get(identifier);
		if (target == null) return;

		if (message.getTarget().getUAL() != null) {
			target.incrementReceived( message.getTarget().getUAL().getHost() );
		}
	}

	public void beginSend(UAN identifier, Message message) {
		ActorProfile target = (ActorProfile)actorProfiles.get(identifier);
		if (target == null) return;    		
		if (message.getSource() != null && target.getUAN().equals(message.getSource().getUAN())) return;

		target.incrementSent( message.getTarget().getUAL().getHost() );
		target.beginSending();
	}
	public void endSend(UAN identifier, Message message) {
		ActorProfile target = (ActorProfile)actorProfiles.get(identifier);
		if (target == null) return;    		
		if (message.getSource() != null && target.getUAN().equals(message.getSource().getUAN())) return;
		target.endSending();
	}

	/*
		The following create new actor profiles and manage their
		entering and exiting the theater
	*/
	public void addProfile(UAN identifier) 		{ actorProfiles.put(identifier, new ActorProfile(identifier)); }
	public void removeProfile(UAN identifier) 	{ actorProfiles.remove(identifier); }

	public void reset() {
		totalProcessed = 0;
		lastResetTime = System.currentTimeMillis();
		stealSources.clear();

		Vector profiles = getActorProfiles();
		for (int i = 0; i < profiles.size(); i++) { ((ActorProfile)profiles.get(i)).reset(); }
	}

	public long totalProcessed() { return processedSinceStart; }

	private Hashtable peersToLatency = new Hashtable();

	protected double getLatencyBetween(int targetLocality, String targetHost, String remoteHost) {
		Double remoteValue = (Double)peersToLatency.get(remoteHost);
		Double targetValue = (Double)peersToLatency.get(targetHost);

		if (targetValue == null) {
			getLatency(targetHost);
			targetValue = (Double)peersToLatency.get(targetHost);
		}

		if (remoteValue == null) {
			getLatency(targetHost);
			targetValue = (Double)peersToLatency.get(targetHost);
		}

		if (targetValue == null) targetValue = new Double(1000);
		if (remoteValue == null) remoteValue = new Double(1000);

		double remoteLatency = remoteValue.doubleValue();
		double targetLatency = targetValue.doubleValue();

		/**
		 *	There are many cases to consider for determining a good measure
		 *	of the latency of two hosts.  We cannot know exactly, but we
		 *	can approximate by knowing what regions they are from this host.
		 *	The only poor misinterpretations occur when the targethost and
		 *	the remotehost are regionally the same distance away - we would
		 *	like to know if they were in the same one.  This would require
		 *	knowing the names of the regions they are in.  mabye for io 0.3
		 */

		if (remoteLatency < 11) {
			return targetLatency;
		} else if (remoteLatency > 10 && remoteLatency < 101) {

			if (targetLocality == 1) return targetLatency;
			else if (targetLocality == 2) {
				if (targetLatency > remoteLatency) return targetLatency;
				else return remoteLatency;
			} else if (targetLocality == 3) return targetLatency;
			else if (targetLocality == 4) return targetLatency;

		} else if (remoteLatency > 100 && remoteLatency < 251) {

			if (targetLocality == 1) return remoteLatency;
			else if (targetLocality == 2) return remoteLatency;
			else if (targetLocality == 3) {
				if (targetLatency > remoteLatency) return targetLatency;
				else return remoteLatency;
			} else if (targetLocality == 4) return targetLatency;

		} else if (remoteLatency > 250) {

			if (targetLocality == 1) return remoteLatency;
			else if (targetLocality == 2) return remoteLatency;
			else if (targetLocality == 3) return remoteLatency;
			else if (targetLocality == 4) {
				if (targetLatency > remoteLatency) return targetLatency;
				else return remoteLatency;
			}
		}

		//targetLocality < 1 or targetLocality > 4
		if (targetLatency < remoteLatency) return remoteLatency;
		else return targetLatency;
	}

	protected double getLatency(String host) {
		Double latency = (Double)peersToLatency.get(host);
		if (latency != null) return latency.doubleValue();

		int port = 22222;
		int pingTrials = 10;

		Socket sensorSocket = null;
		ObjectInputStream ois = null;  
		ObjectOutputStream oos = null;

		// Opening a socket with the delay sensor at target Peer
		try {
			sensorSocket = new Socket(host,port);

			//disable the nagle algorithm since we are sending only one byte
			sensorSocket.setTcpNoDelay(false);
			oos = new ObjectOutputStream(sensorSocket.getOutputStream());
		} catch (UnknownHostException e) {
			e.printStackTrace();
		} catch (IOException e) {
			e.printStackTrace();
		}

		// Ping several times
		long totalTime = 0;
		for (int i = 0; i < pingTrials; i++) {
			long startTime = System.currentTimeMillis();
			try {
				oos.write(0);
				oos.flush();
				if (ois == null) ois = new ObjectInputStream(sensorSocket.getInputStream() );

				//Read the byte back
				ois.readByte();
			} catch (IOException e1) {
				e1.printStackTrace();
			}
			if ( i!= 0) totalTime += System.currentTimeMillis() - startTime;
		}

		try {
			oos.close();
			ois.close();
			sensorSocket.close();
		} catch (IOException e1) {
			e1.printStackTrace();
		}

		if ( System.getProperty("silent") == null) {
			System.err.println("LATENCY: " + host + " at " + (totalTime/(pingTrials-1)));
		}

		peersToLatency.put(host, new Double(totalTime/(pingTrials-1)));
		return totalTime/(pingTrials-1);
	}
}
